/**
 * accesspoint.c -- AP implementation
 * 
 * Copyright (C) 2011 Pantelis A. Frangoudis <pfrag@aueb.gr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "accesspoint.h"

/**
 * A report by an AP -- always assumed truthful
 * CHANGED: CAUTION: In some scenarios, the AP does not report. In these cases
 * ap_report should "relax" edge weights
 * HARDCODED: AP report weight == 1!! -- changed from '2' (5/5/2011)
 * ap_edges == 0: Ignore AP-scanned edges in the IG
 * ap_edges == 1 && ap_report == 0: Remove AP-scanned edges from IG
 * ap_edges == 1 && ap_report == 1: 5/5/2011 version
 */

void ap_report(AP *ap, int ap_edges, int ap_report) {
	int i = 0;
	int j = 0;
	int from = 0;
	int to = 0;

	//no nearby APs
	if (ap->scan_list_size <= 0) {
		return;
	}

	if (!ap_edges) {
		return;
	}

	if (!ap_report || !ap->is_managed) {
		for (i = 0; i < ap->scan_list_size; i++) {
			from = ap->scan_list[i]->id;

			//self<-->scanned edges
			/* 6/5/2011 */
			if (ap->is_managed || ap->scan_list[i]->is_managed) {
				/* 16/5/11: akyro sxolio: An to reporting AP einai unmanaged kai h akmh perilamvanei managed, h akmh tha ginei relax. An to weight me th relax ginei 0, tote h akmh tha afairethei. Ston 1o gyro, client->r == 0. An loipon kapoioi clients tin exoun kanei report, h plhroforia ayti tha xathei prin klhthei h filter!! Gi'ayto, prepei na "thymomaste" oti oi clients thn exoun kanei rep, kati pou kanoume vevaia otan sthn client->report */					
				relax_edge(IG, from, ap->id, AP_WEIGHT);
				relax_edge(IG, ap->id, from, AP_WEIGHT);
			}

			for (j = i; j < ap->scan_list_size; j++) {
				if (i == j) continue;
				to = ap->scan_list[j]->id;

				//watch out!! we do stuff on the *real* IG (there's no "reported" one)
				if (ap->scan_list[i]->is_managed || ap->scan_list[j]->is_managed) {
					relax_edge(IG, from, to, AP_WEIGHT);
					relax_edge(IG, to, from, AP_WEIGHT);
				}
			}
		}
	}

	//else, do nothing (for now...)
}

/**
 * Dealloc mem for an AP
 */
void free_ap(AP* ap) {
	if (ap && (ap->num_assoc > 0) ) {
		free(ap->assoc_list);
	}

	if (ap && (ap->scan_list_size > 0) ) {
		free(ap->scan_list);
	}

	free(ap);
}

void flush_ap_info(AP *ap) {
	if (ap && (ap->num_assoc > 0) ) {
		free(ap->assoc_list);
		ap->assoc_list = NULL;
		ap->num_assoc = 0;
	}


	if (ap && (ap->scan_list_size > 0) ) {
		free(ap->scan_list);
		ap->scan_list = NULL;
		ap->scan_list_size = 0;
	}

}

/**
 * AP ap scans all APs in blocks[bid]
 * Returns a ptr to the closest AP and sets min_distance
 * [or returns NULL if no AP is in range]
 */
AP* ap_scan_block(AP *ap, int bid, double *min_distance) {
	//Check whether bid is a "valid" block
	//normally, each client needs to scan 9 blocks. If he's on the
	//boundary, this has to be detected so that he doesn't scan 
	//faraway blocks

	//TODO: no scan at wrap-around blocks!!

	if (bid < 0 || bid >= ((int)(terrainX/range) + 1) * ((int)(terrainY/range) + 1) ) {
		//"invalid" block
		return NULL;
	}

	Block *b = &(blocks[bid]);
	int i = 0;
	double distance;
	AP *retval = NULL;

	for (i = 0; i < blocks[bid].apnum; i++) {
		if (ap->id == b->aps[i]->id) continue;
		//printf("Client %d scanning block %d (%d APs), AP %d\n", c->id, bid, blocks[bid].apnum, i);
		distance = sqrt(pow(ap->x - b->aps[i]->x, 2) +  pow(ap->y - b->aps[i]->y, 2));

		//if in range, add aps[i] to scan list
		if (distance <= range) {
			if (ap->scan_list_size == 0) {
				ap->scan_list = (AP**)malloc(sizeof(AP*));
			}
			else {
				ap->scan_list = (AP**)realloc( ap->scan_list, (ap->scan_list_size + 1)*sizeof(AP*));
			}
			ap->scan_list[ap->scan_list_size] = b->aps[i];
			ap->scan_list_size++;
		}

		if (distance < *min_distance && distance < range) {
			*min_distance = distance;
			retval = b->aps[i];
		}
	}

	return retval;
}

void ap_scan(AP *ap) {
#ifdef USE_BLOCKS
	//get block
	int bid = get_block(ap->x, ap->y);

	double min_distance = 10000000000.0;
	double temp_min_distance = min_distance;

	AP *closest_ap = NULL;
	AP *temp_ap = NULL;

	//scan blocks
	temp_ap = ap_scan_block(ap, bid - 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = ap_scan_block(ap, bid, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = ap_scan_block(ap, bid + 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = ap_scan_block(ap, bid - (int)terrainX/range - 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = ap_scan_block(ap, bid - (int)terrainX/range, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = ap_scan_block(ap, bid - (int)terrainX/range + 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = ap_scan_block(ap, bid + (int)terrainX/range - 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = ap_scan_block(ap, bid + (int)terrainX/range, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = ap_scan_block(ap, bid + (int)terrainX/range + 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}
#else
	int i = 0;
	double distance;

	for (i = 0; i < num_aps; i++) {
		if (ap->id == aps[i]->id) continue;

		distance = sqrt(pow(ap->x - aps[i]->x, 2) +  pow(ap->y - aps[i]->y, 2));

		//if in range, add aps[i] to scan list
		if (distance <= range) {
			if (ap->scan_list_size == 0) {
				ap->scan_list = (AP**)malloc(sizeof(AP*));
			}
			else {
				ap->scan_list = (AP**)realloc( ap->scan_list, (ap->scan_list_size + 1)*sizeof(AP*));
			}
			ap->scan_list[ap->scan_list_size] = aps[i];
			ap->scan_list_size++;
		}
	}

#endif

}

void ap_set_behaviour(AP *ap, attack_scenario *as) {
	if (rand_double(1) <= as->managed_ap_ratio) {
		ap->is_managed = 1;
	}
	else {
		ap->is_managed = 0;
	}
}

