/**
 * client.c -- Client implementation
 * 
 * Copyright (C) 2011 Pantelis A. Frangoudis <pfrag@aueb.gr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "client.h"

/**
 * Scan and associate with random *managed* (5/5/11) AP
 */
void associate(Client *client) {
	double minDistance = 10000000000.0;
	double distance;

	int i = 0;

#ifdef USE_BLOCKS
	/* build scan list -- ignore return value (closest ap) */
	AP *selected_ap = NULL;
	int has_managed = 0;
	if (scan(client)) {
		/* Check if client has a managed AP in range */
		for (i = 0; i < client->scan_list_size; i++) {
			/* yes, there is at least one managed AP in range */
			if (client->scan_list[i]->is_managed) {
				has_managed = 1;
				break;
			}
		}

		if (!has_managed) {
			/* no managed APs -- random association*/
			selected_ap = client->scan_list[(int)rand_double(client->scan_list_size)];

		}
		else {

			/* Managed APs are in range -- loop until you find a managed one to assoc with */
			while (1) {
				i = (int)rand_double(client->scan_list_size);
					selected_ap = client->scan_list[i];
				if (client->scan_list[i]->is_managed) {
					break;
				}
			}
		}

		if (selected_ap->num_assoc == 0) {
			selected_ap->assoc_list = (Client**)malloc(sizeof(Client*));
		}
		else {
			selected_ap->assoc_list = (Client**)realloc( selected_ap->assoc_list, (selected_ap->num_assoc + 1)*sizeof(Client*));
		}
		selected_ap->assoc_list[selected_ap->num_assoc] = client;
		selected_ap->num_assoc++;
		client->assoc_ap = selected_ap;
	}
	else {
		/* no APs in range */
		client->assoc_ap = NULL;
	}
#else 
	//find closest AP
	int closestAP = -1;

	for (i = 0; i < num_aps; i++) {
		distance = sqrt(pow(client->x - aps[i]->x, 2) +  pow(client->y - aps[i]->y, 2));

		//if in range, add aps[i] to scan list
		if (distance <= range) {
			if (client->scan_list_size == 0) {
				client->scan_list = (AP**)malloc(sizeof(AP*));
			}
			else {
				client->scan_list = (AP**)realloc( client->scan_list, (client->scan_list_size + 1)*sizeof(AP*));
			}
			client->scan_list[client->scan_list_size] = aps[i];
			client->scan_list_size++;
		}

		//if this ap is closer and the client is within range...
		if (distance < minDistance && distance < range) {
			minDistance = distance;
			closestAP = i;
		}
	}

	/* 5/5/11: select a random AP from the list [ignore closestAP stuff] */
	if (closestAP >= 0) {
		i = (int)rand_double(client->scan_list_size);

		if (aps[i]->num_assoc == 0) {
			aps[i]->assoc_list = (Client**)malloc(sizeof(Client*));
			//memset(aps[i]->assoc_list, 0, sizeof(Client*));
		}
		else {
			aps[i]->assoc_list = (Client**)realloc( aps[i]->assoc_list, (aps[i]->num_assoc + 1)*sizeof(Client*));
			//memset(aps[i]->assoc_list[aps[i]->num_assoc], 0, sizeof(Client*));
		}
		aps[i]->assoc_list[aps[i]->num_assoc] = client;
		aps[i]->num_assoc++;
		client->assoc_ap = aps[i];
	}
	else {
		client->assoc_ap = NULL;
	}
#endif

}

/**
 * Client c scans all APs in blocks[bid]
 * Returns a ptr to the closest AP and sets min_distance
 * [or returns NULL if no AP is in range]
 */
AP* scan_block(Client *c, int bid, double *min_distance) {
	//Check whether bid is a "valid" block
	//normally, each client needs to scan 9 blocks. If he's on the
	//boundary, this has to be detected so that he doesn't scan 
	//faraway blocks

	//TODO: no scan at wrap-around blocks!!

	if (bid < 0 || bid >= ((int)(terrainX/range) + 1) * ((int)(terrainY/range) + 1) ) {
		//"invalid" block
		return NULL;
	}

	Block *b = &(blocks[bid]);
	int i = 0;
	double distance;
	AP *retval = NULL;

	for (i = 0; i < blocks[bid].apnum; i++) {
		//printf("Client %d scanning block %d (%d APs), AP %d\n", c->id, bid, blocks[bid].apnum, i);
		distance = sqrt(pow(c->x - b->aps[i]->x, 2) +  pow(c->y - b->aps[i]->y, 2));

		//if in range, add aps[i] to scan list
		if (distance <= range) {
			if (c->scan_list_size == 0) {
				c->scan_list = (AP**)malloc(sizeof(AP*));
			}
			else {
				c->scan_list = (AP**)realloc( c->scan_list, (c->scan_list_size + 1)*sizeof(AP*));
			}
			c->scan_list[c->scan_list_size] = b->aps[i];
			c->scan_list_size++;
		}

		//if this ap is closer and the client is within range...
		if (distance < *min_distance && distance < range) {
			*min_distance = distance;
			retval = b->aps[i];
		}
	}

	return retval;
}

AP* scan(Client *c) {
	//get client block
	int bid = get_block(c->x, c->y);

	double min_distance = 10000000000.0;
	double temp_min_distance = min_distance;

	AP *closest_ap = NULL;
	AP *temp_ap = NULL;

	//scan blocks
	temp_ap = scan_block(c, bid - 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = scan_block(c, bid, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = scan_block(c, bid + 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = scan_block(c, bid - (int)terrainX/range - 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = scan_block(c, bid - (int)terrainX/range, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = scan_block(c, bid - (int)terrainX/range + 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = scan_block(c, bid + (int)terrainX/range - 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = scan_block(c, bid + (int)terrainX/range, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	temp_ap = scan_block(c, bid + (int)terrainX/range + 1, &temp_min_distance);
	if (temp_ap && temp_min_distance < min_distance) {
		closest_ap = temp_ap;
		min_distance = temp_min_distance;
	}
	else {
		temp_min_distance = min_distance;
	}

	return closest_ap;
}

/**
 * Configure client behavior.
 * Cases (changed -- 5/5/2011):
 * - Truthful (c->defection_probability = 0)
 * - Potential Attacker
 *	(c->defection_probability determines whether he will
 *   attack at round i)
 */
void set_strategy(Client *c, attack_scenario *as) {
	//if not, is he an attacker?
	if (rand_double(1) <= as->attacker_ratio) {
		c->defection_probability = as->defection_probability;
	}
	else {
		//truthful but non-trusted
		c->defection_probability = 0.0;
	}
}

/**
 * Executes an attack 
 * [For the moment, no collusion. To fix it, on round initialization we should 
 * decide, for each client, if he'll defect or not beforehand.]
 */
void attack(Client *c) {
	int i = 0;
	int j = 0;

	c->will_attack = 0;

	//not assoc'ed clients can't attack
	if (c->assoc_ap == NULL) {
		return;
	}

	//now, decide if he'll attack or not
	if (rand_double(1) < c->defection_probability) {
		//ATTACK!
		c->will_attack = 1;
	}

}

/**
 * Client c reports [adds IG edges]
 */
void report(Client *c, int do_report) {
	int i = 0;
	int j = 0;
	int from = 0;
	int to = 0;
	int roamers = 0;
	double weight = 0;

	//Client not associated with anyone --> no APs in range, no report
	if (c->assoc_ap == NULL) {
		return;
	}

	//report weight is equal to client's reputation
	weight = c->r;

	//super poustia: ta reports "peirazoun" apeytheias to true IG, afairwntas
	//apo kathe akmi to varos stin periptwsi pou o client einai attacker
	//'h mh kanontas tipota stin periptwsi pou o cli einai truthful
	//[ayta proypothetoun allages sto pws leitourgei o simulator]
//	if (!c->truthful || !do_report) {
		for (i = 0; i < c->scan_list_size; i++) {
			from = c->scan_list[i]->id;
			for (j = i; j < c->scan_list_size; j++) {
				if (i == j) continue;
				to = c->scan_list[j]->id;

				/* 5/5/11: if a cli is assoced with unmanaged ap, he cannot report! */
				//if ( (!c->trusted && c->will_attack) || !do_report) {
				if (c->will_attack || !do_report || !c->assoc_ap->is_managed) {
					//watch out!! we do stuff on the *real* IG (there's no "reported" one)
					if (c->scan_list[i]->is_managed || c->scan_list[j]->is_managed) {
						relax_edge(IG, from, to, weight);
						relax_edge(IG, to, from, weight);
					}
				}
				else {
					//else (truthful), keep track of user reporting an edge
					if (c->scan_list[i]->is_managed || c->scan_list[j]->is_managed) {
						add_reporter(IG, from, to, c);
						add_reporter(IG, to, from, c);
						c->reported_edges += 2;
						if (is_type1(c->scan_list[i], c->scan_list[j], range)) {
							c->reportedT1++;
						}
						else {
							c->reportedT2++;
						}
					}
				}
			}
		}
//	}
}

void print_fake_list(Client *c) {
	int i = 0;
	fprintf(stderr, "Attacker %d: ", c->id);
	for (i = 0; i < c->fake_list_size; i++) {
		fprintf(stderr, "%d\t", c->fake_list[i]->id);
	}
	fprintf(stderr, "\n");
}

void print_scan_list(Client *c) {
	int i = 0;
	fprintf(stderr, "Client %d [willattack: %d, assoc_ap: %d, scan_list_size: %d]: ", c->id, c->will_attack, (c->assoc_ap?c->assoc_ap->id:-1), c->scan_list_size);
	for (i = 0; i < c->scan_list_size; i++) {
		fprintf(stderr, "%d[%d]\t", c->scan_list[i]->id, c->scan_list[i]->is_managed);
	}
	fprintf(stderr, "\n");
}

/**
 * Free mem of a client struct
 */
void free_client(Client *client) {
	if (client->fake_list_size && client->fake_list) {
		free(client->fake_list);
	}
	if (client->scan_list_size && client->scan_list) {
		free(client->scan_list);
	}
	free(client);
}

 void flush_client_info(Client *client) {
	if (client->fake_list_size && client->fake_list) {
		free(client->fake_list);
		client->fake_list = NULL;
	}
	if (client->scan_list_size && client->scan_list) {
		free(client->scan_list);
		client->scan_list = NULL;
	}
}
