/**
 * gui.c -- Minimal GUI-based demonstration of the simulator terrain. Shows
 * where APs and clients are placed and cell range.
 * 
 * Copyright (C) 2011 Pantelis A. Frangoudis <pfrag@aueb.gr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#include "gui.h"

void get_color_from_theme(int color, GdkColor *gdkc) {
	gdkc->pixel = -1;
	GdkColormap *map = gdk_colormap_get_system();
	switch(color) {
		case COLOR_RED:
			gdk_color_parse("#ff0000", gdkc);
			break;

		case COLOR_BLUE:
			gdk_color_parse("#0000ff", gdkc);
			break;

		case COLOR_GREEN:
			gdk_color_parse("#00ff00", gdkc);
			break;

		case COLOR_WHITE:
			gdk_color_parse("#ffffff", gdkc);
			break;

		case COLOR_YELLOW:
			gdk_color_parse("#ffff00", gdkc);
			break;

		case COLOR_CYAN:
			gdk_color_parse("#00ffff", gdkc);
			break;

		case COLOR_MAGENDA:
			gdk_color_parse("#ff00ff", gdkc);
			break;

		case COLOR_BLACK: default:
			gdk_color_parse("#000000", gdkc);
			break;
	}	
	gdk_colormap_alloc_color(map, gdkc, FALSE, TRUE);
}


/**
 * Draw an AP
 */
void draw_ap(theme *t, gdouble x, gdouble y, int apid, int channel, int is_managed) {
	GdkGC *gc = gdk_gc_new(GDK_DRAWABLE(pixmap));

	GdkColor col;
	if (is_managed) {
		get_color_from_theme(t->managed_ap_color, &col);
	}
	else {
		get_color_from_theme(t->ap_color, &col);
	}

	gdk_gc_set_foreground(gc, &col);

	GdkRectangle update_rect;
	update_rect.x = x - t->ap_size/2;
	update_rect.y = y - t->ap_size/2;
	update_rect.width = t->ap_size;
	update_rect.height = t->ap_size;

	GdkPoint points[3];
	
	switch(t->ap_shape) {
		case SHAPE_CIRCLE:
			gdk_draw_arc(
				pixmap,
				gc,
				TRUE,
				update_rect.x,
				update_rect.y,
				update_rect.width,
				update_rect.height,
				0,
				360*64);
			break;

		case SHAPE_TRI:
			points[0].x = (gint16)x;
			points[0].y = (gint16)y - t->ap_size/2;
			points[1].x = (gint16)x - t->ap_size/2;
			points[1].y = (gint16)y + t->ap_size/2;
			points[2].x = (gint16)x + t->ap_size/2;
			points[2].y = (gint16)y + t->ap_size/2;
			gdk_draw_polygon(pixmap, gc, TRUE, points, 3);
			break;

		case SHAPE_CROSS:
			gdk_draw_line(pixmap, gc, (gint)x - t->ap_size/2, (gint)y, (gint)x + t->ap_size/2, (gint)y);
			gdk_draw_line(pixmap, gc, (gint)x, (gint)y - t->ap_size/2, (gint)x, (gint)y + t->ap_size/2);
			break;

		case SHAPE_RECT: default:
			gdk_draw_rectangle (pixmap, gc, TRUE,
					update_rect.x, update_rect.y,
					update_rect.width, update_rect.height);
			break;
	}

	//draw AP ID [assumed integer]
	if (t->show_ap_id) {
		char str_ap_id[10];
		memset(str_ap_id, 0, 10);
		sprintf(str_ap_id, "%d", apid);
	
		GdkScreen *screen = gdk_drawable_get_screen (pixmap);
		PangoRenderer *renderer = gdk_pango_renderer_get_default (screen);
		
		PangoContext *context = gdk_pango_context_get_for_screen (screen);
		PangoLayout *layout = pango_layout_new (context);
	
		pango_layout_set_text (layout, str_ap_id, -1);
	
		gdk_pango_renderer_set_drawable (GDK_PANGO_RENDERER (renderer), pixmap);
	
		GdkGC *gc2 = gdk_gc_new (pixmap);
		gdk_pango_renderer_set_gc (GDK_PANGO_RENDERER (renderer), gc2);
	
		//channel color should be visible	

		GdkColor color;
		get_color_from_theme(COLOR_RED, &color);
	
		gdk_pango_renderer_set_override_color (
			GDK_PANGO_RENDERER (renderer),
			PANGO_RENDER_PART_FOREGROUND, &color);
	
		pango_layout_context_changed (layout);
		gdk_draw_layout (pixmap, gc2, x - t->ap_size/2, y - 2*t->ap_size, layout);
	
		gdk_pango_renderer_set_override_color (GDK_PANGO_RENDERER (renderer),
						PANGO_RENDER_PART_FOREGROUND, NULL);
		gdk_pango_renderer_set_drawable (GDK_PANGO_RENDERER (renderer), NULL);
		gdk_pango_renderer_set_gc (GDK_PANGO_RENDERER (renderer), NULL);
	

		g_object_unref (layout);
		g_object_unref (context);
		g_object_unref (gc2);
	}
	
	//draw channel number
	if (t->show_channels) {
		char str_cnl[3];
		memset(str_cnl, 0, 3);
		sprintf(str_cnl, "%d", channel);
	
		GdkScreen *screen = gdk_drawable_get_screen (pixmap);
		PangoRenderer *renderer = gdk_pango_renderer_get_default (screen);
		
		PangoContext *context = gdk_pango_context_get_for_screen (screen);
		PangoLayout *layout = pango_layout_new (context);
	
		pango_layout_set_text (layout, str_cnl, -1);
	
		gdk_pango_renderer_set_drawable (GDK_PANGO_RENDERER (renderer), pixmap);
	
		GdkGC *gc2 = gdk_gc_new (pixmap);
		gdk_pango_renderer_set_gc (GDK_PANGO_RENDERER (renderer), gc2);
	
		//channel color should be visible	
		GdkColor color;
		color.red = 0;
		color.green = 0;
		color.blue = 0;
	
		gdk_pango_renderer_set_override_color (
			GDK_PANGO_RENDERER (renderer),
			PANGO_RENDER_PART_FOREGROUND, &color);
	
		pango_layout_context_changed (layout);
		gdk_draw_layout (pixmap, gc2, x, y, layout);
	
		gdk_pango_renderer_set_override_color (GDK_PANGO_RENDERER (renderer),
						PANGO_RENDER_PART_FOREGROUND, NULL);
		gdk_pango_renderer_set_drawable (GDK_PANGO_RENDERER (renderer), NULL);
		gdk_pango_renderer_set_gc (GDK_PANGO_RENDERER (renderer), NULL);
	
		g_object_unref (layout);
		g_object_unref (context);
		g_object_unref (gc2);
	}

	g_object_unref(gc);
}

/**
 * Draw a client
 */
void draw_client(theme *t, gdouble x, gdouble y) {
	GdkGC *gc = gdk_gc_new(GDK_DRAWABLE(pixmap));

	GdkColor col;
	get_color_from_theme(t->cli_color, &col);
	gdk_gc_set_foreground(gc, &col);

	GdkRectangle update_rect;
	update_rect.x = x - t->cli_size/2;
	update_rect.y = y - t->cli_size/2;
	update_rect.width = t->cli_size;
	update_rect.height = t->cli_size;

	GdkPoint points[3];
	
	switch(t->cli_shape) {
		case SHAPE_CIRCLE:
			gdk_draw_arc(
				pixmap,
				gc,
				TRUE,
				update_rect.x,
				update_rect.y,
				update_rect.width,
				update_rect.height,
				0,
				360*64);
			break;

		case SHAPE_TRI:
			points[0].x = (gint16)x;
			points[0].y = (gint16)y - t->cli_size/2;
			points[1].x = (gint16)x - t->cli_size/2;
			points[1].y = (gint16)y + t->cli_size/2;
			points[2].x = (gint16)x + t->cli_size/2;
			points[2].y = (gint16)y + t->cli_size/2;
			gdk_draw_polygon(pixmap, gc, TRUE, points, 3);
			break;

		case SHAPE_CROSS:
			gdk_draw_line(pixmap, gc, (gint)x - t->cli_size/2, (gint)y, (gint)x + t->cli_size/2, (gint)y);
			gdk_draw_line(pixmap, gc, (gint)x, (gint)y - t->cli_size/2, (gint)x, (gint)y + t->cli_size/2);
			break;

		case SHAPE_RECT: default:
			gdk_draw_rectangle (pixmap, gc, TRUE,
					update_rect.x, update_rect.y,
					update_rect.width, update_rect.height);
			break;
	}

	g_object_unref(gc);
}

void draw_radius(theme *t, gdouble x, gdouble y, double radius) {
	GdkGC *gc = gdk_gc_new(GDK_DRAWABLE(pixmap));

	GdkColor col;
	get_color_from_theme(t->rad_color, &col);
	gdk_gc_set_foreground(gc, &col);

	int i = 0;
	gdk_draw_arc(
		pixmap,
		gc,
		FALSE,
		x - radius,
		y - radius,
		2 * radius,
		2 * radius,
		0,
		360*64);
}

/**
 * Draw the network topology
 */
void draw_topology(GtkWidget *widget, simsettings *s) {
	if (pixmap) {
		g_object_unref (pixmap);
	}

	pixmap = gdk_pixmap_new (widget->window,
				widget->allocation.width,
				widget->allocation.height,
				-1);

	gdk_draw_rectangle (pixmap,
			widget->style->white_gc,
			TRUE,
			0, 0,
			widget->allocation.width,
			widget->allocation.height);

	double viewer_width = widget->allocation.width;
	double viewer_height = widget->allocation.height;

	//draw APs
	int i = 0;
	for (i = 0; i < num_aps; i++) {
		draw_ap(&(s->theme), s->aps[i]->x * viewer_width / terrainX, s->aps[i]->y * viewer_height / terrainY, aps[i]->id, aps[i]->channel, aps[i]->is_managed);
		if (s->theme.show_radius) {
			draw_radius(&(s->theme), s->aps[i]->x * viewer_width / s->terrainX, s->aps[i]->y * viewer_height / s->terrainY, s->radius* viewer_height / s->terrainY);
		}
	}

	//draw clients
	for (i = 0; i < s->num_clients; i++) {
		draw_client(&(s->theme), s->clients[i]->x * viewer_width / s->terrainX, s->clients[i]->y * viewer_height / s->terrainY);
	}

	//draw edges
	gint fromX, fromY, toX, toY;
	int j;
	GdkGC *gc = gdk_gc_new(GDK_DRAWABLE(pixmap));
	if (s->theme.show_edges) {
		for (i = 0; i < s->num_aps; i++) {
			for (j = i; j < s->num_aps; j++) {
				if (
					sqrt( pow(s->aps[i]->x - s->aps[j]->x, 2) + pow(s->aps[i]->y - s->aps[j]->y, 2) ) <= 2*s->radius && 
					sqrt( pow(s->aps[i]->x - s->aps[j]->x, 2) + pow(s->aps[i]->y - s->aps[j]->y, 2) > s->radius)
					)  {
					if (s->aps[i]->is_managed || s->aps[j]->is_managed) {
						fromX = (gint)(s->aps[i]->x * viewer_width / s->terrainX);
						fromY = (gint)(s->aps[i]->y * viewer_height / s->terrainY);
						toX = (gint)(s->aps[j]->x * viewer_width / s->terrainX);
						toY = (gint)(s->aps[j]->y * viewer_height / s->terrainY);

						gdk_draw_line(pixmap, gc, fromX, fromY, toX, toY);
					}
				}
			}
		}
	}
}


/**
 * Create a new backing pixmap of the appropriate size 
 */
gboolean configure_event(GtkWidget *widget, GdkEventConfigure *event) {
	if (pixmap) {
		g_object_unref (pixmap);
	}

	pixmap = gdk_pixmap_new (widget->window,
				widget->allocation.width,
				widget->allocation.height,
				-1);
	gdk_draw_rectangle (pixmap,
			widget->style->white_gc,
			TRUE,
			0, 0,
			widget->allocation.width,
			widget->allocation.height);
	
	return TRUE;
}

/**
 * Handling expose events. All points are redrawn
 */
gboolean expose_event(GtkWidget *widget, GdkEventExpose *event, void *data)  {
	simsettings *s = (simsettings*)data;

	draw_topology(widget, s);

	gdk_draw_drawable (
		widget->window,
		widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		pixmap,
		event->area.x, event->area.y,
		event->area.x, event->area.y,
		event->area.width, event->area.height);

	return FALSE;
}

int show_gui(int argc, char **argv, simsettings *s) {
	GtkWidget *window;
	GtkWidget *drawing_area;
	GtkWidget *vbox;

	pixmap = NULL;

	gtk_init (&argc, &argv);

	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_name (window, "Terrain view");

	vbox = gtk_vbox_new (TRUE, 0);
	gtk_container_add (GTK_CONTAINER (window), vbox);
	gtk_widget_show (vbox);

	//Create the drawing area
	drawing_area = gtk_drawing_area_new ();
	gtk_widget_set_size_request (GTK_WIDGET(drawing_area), 500, 500);

	gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);

	g_signal_connect (drawing_area, "expose_event", G_CALLBACK (expose_event), (void*)s);
//	g_signal_connect (drawing_area, "configure_event", G_CALLBACK (configure_event), NULL);

//	gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK| GDK_LEAVE_NOTIFY_MASK);
	
	gtk_widget_show (drawing_area);
	gtk_widget_show (window);

	gtk_main ();

	return 0;
}
