/**
 * ig.c -- Coverage graph implementation. Each vertex is an AP and an edge 
 * represents overlapping coverage between two cells. The graph is implemented
 * as an adjacency list. For each edge, we keep track of those who reported it.
 * 
 * Copyright (C) 2011 Pantelis A. Frangoudis <pfrag@aueb.gr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "ig.h"

Node *init_ig(AP **aps, int num_nodes) {
	int i = 0;

	Node *retval = (Node*)malloc(num_nodes*sizeof(Node));
	memset(retval, 0, num_nodes*sizeof(Node));
	for (i = 0; i < num_nodes; i++) {
		retval[i].ap = aps[i];
		retval[i].neighbors = NULL;
	}

	return retval;
}

Node *init_reported_ig(AP **aps, AP **fake_aps, int num_real_nodes, int num_fake_nodes) {
	int i = 0;

	Node *retval = (Node*)malloc( (num_real_nodes + num_fake_nodes)*sizeof(Node));
	memset(retval, 0, (num_real_nodes + num_fake_nodes)*sizeof(Node));
	for (i = 0; i < num_real_nodes; i++) {
		retval[i].ap = aps[i];
		retval[i].neighbors = NULL;
	}

	for (i = num_real_nodes; i < (num_real_nodes + num_fake_nodes); i++) {
		retval[i].ap = fake_aps[i - num_real_nodes];
		retval[i].neighbors = NULL;
	}

	return retval;
}

void free_ig(Node *ig, int num_nodes) {
	int i = 0;
	for (i = 0; i < num_nodes; i++) {
		//for each node, first free edge list
		Edge *cur = ig[i].neighbors;
		Edge *todelete = NULL;
		while (cur != NULL) {
			todelete = cur;
			cur = cur->next;

			if (todelete->reporters != NULL) {
				free(todelete->reporters);
			}

			free(todelete);
		}
	}

	//free IG
	free(ig);
}

/**
 * Add an edge to the graph. If the edge exists, update its weight
 * Returns 1 if new edge was added, 0 otherwise
 */
int add_edge(Node *ig, int from, int to, double weight, int by_truthful) {
	Edge *edge = ig[from].neighbors;
	Edge *prev = edge;

	while (edge != NULL) {
		if (edge->to->ap->id == to) {
			edge->weight += weight;
			if (by_truthful) edge->reported_by_truthful = 1;
			return 0;
		}
		prev = edge;
		edge = edge->next;
	}

	//edge not found --> add new one
	/* 15/5: note: we may add a zero-weight edge */
	Edge *newedge = (Edge*)malloc(sizeof(Edge));
	memset(newedge, 0, sizeof(Edge));

	newedge->num_reporters = 0;
	newedge->reporters = NULL;

	newedge->to = &ig[to];
	newedge->weight = weight;
	newedge->next = NULL;
	if (by_truthful) newedge->reported_by_truthful = 1;
	else newedge->reported_by_truthful = 0;

	if (!prev) {
		ig[from].neighbors = newedge;
	}
	else {
		prev->next = newedge;
	}
	return 1;
}

/**
 * Remove an edge
 */
void remove_edge(Node *ig, int from, int to) {
	Edge *cur = ig[from].neighbors;
	Edge *prev = NULL;
	while (cur != NULL) {
		if (cur->to->ap->id == to) {
			//edge found

			//remove edge from IG
			if (prev == NULL) {
				ig[from].neighbors = cur->next;
				if (cur->reporters != NULL) {
					free(cur->reporters);
				}
				free(cur);
				cur = ig[from].neighbors;
			}
			else {
				prev->next = cur->next;
				if (cur->reporters != NULL) {
					free(cur->reporters);
				}
				free(cur);
				cur = prev->next;
			}
			return;
		}
		prev = cur;
		cur = cur->next;
	}
}

/**
 * Searches for edge and reduces its weight by "weight". If it reaches 0,
 * edge is removed
 */
void relax_edge(Node *ig, int from, int to, double weight) {
	Edge *cur = ig[from].neighbors;
	Edge *prev = NULL;
	while (cur != NULL) {
		if (cur->to->ap->id == to) {
			//edge found
			cur->weight -= weight;

			if (cur->weight <= 0 && !cur->reported_by_truthful) {
//			if (cur->weight <= 0) {
				//remove edge from IG if its weight reaches 0
				if (prev == NULL) {
					ig[from].neighbors = cur->next;
					if (cur->reporters != NULL) {
						free(cur->reporters);
					}
					free(cur);
					cur = ig[from].neighbors;
				}
				else {
					prev->next = cur->next;
					if (cur->reporters != NULL) {
						free(cur->reporters);
					}
					free(cur);
					cur = prev->next;
				}
			}
			return;
		}
		prev = cur;
		cur = cur->next;
	}
}

/**
 * Returns the weight of an edge or 0 if it doesn't exist
 */
double edge(Node *ig, int from, int to) {
	Edge *edge = ig[from].neighbors;
	while (edge != NULL) {
		if (edge->to->ap->id == to) {
			return edge->weight;
		}
		edge = edge->next;
	}
	return 0;
}

void add_reporter(Node *ig, int from, int to, Client *c) {
	Edge *edge = ig[from].neighbors;
	while (edge != NULL) {
		if (edge->to->ap->id == to) {
			//edge found
			edge->num_reporters++;
			if (edge->num_reporters == 1) {
				edge->reporters = (Client**)malloc(sizeof(Client*));
				memset(edge->reporters, 0, sizeof(Client*));
				edge->reporters[0] = c;
			}
			else {
				edge->reporters = (Client**)realloc(edge->reporters, edge->num_reporters * sizeof(Client*));
				edge->reporters[edge->num_reporters - 1] = c;
			}
			return;
		}
		edge = edge->next;
	}
}

