/**
 * simconfig.c -- Configuration file parser. It parses files with the following 
 * format: 
 * 
 * # comment
 * key1 value1
 * key2 value2
 * ....
 *
 * Keys are case-insensitive.
 *
 * Copyright (C) 2011 Pantelis A. Frangoudis <pfrag@aueb.gr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "simconfig.h"

int parse_line(char *line, char *key, char *value) {
	int len = strlen(line);

	int i = 0;
	int j = 0;
	int k = 0;

	/* removes space from the beginning of line */
	for (i=0; i<len; i++) {
		if ( !isspace(line[i]) ) {
			break;
		}
	}

	/* if its a comment ---> do nothing */
	if (line[i] == '#') {
		return 0;
	}

	/* if its an empty line, do nothing */
	if (line[i] == '\0') {
		return 0;
	}

	/* gets key */
	for (j=i; j<len; j++) {
		if ( isspace(line[j]) ) {
			break;
		}
		key[k] = line[j];
		k++;
	}

	/* removes trailing space */
	while (isspace(line[j]) && j<len) {
		j++;
	}

	/* key without value */
	if (j == len) {
		*value = '\0';
		return 1;
	}

	/* the rest of it is the value */
	strncpy(value, line + j, len - j);

	/* remove trailing whitespace from value (replace them with \0) */
	for (i=(strlen(value) - 1); i>=0; i--) {
		if (isspace(value[i])) {
				value[i] = '\0';
		}
		else {
				break;
		}
	}
	return 2;
}

int parse_conffile(char *fname, char **confoptions, char **confvalues, int optlen) {
	FILE *fp;
	char *line;
	char key[80];
	char value[80];
	size_t linelen;
	int i;
	int ret;

	fp = fopen(fname, "r");
	if (!fp) {
		return -1;
	}

	linelen = 0;
	line = NULL;
	while ( (linelen = getline(&line, &linelen, fp)) != -1) {
		if (!line) continue;

		/* remove newline character from line */
		for (i = linelen - 1; i >= 0; i--) {
			if (line[i] != '\r' && line[i] != '\n') break;
			line[i] = '\0';
		}

		memset(key, 0, 80);
		memset(value, 0, 80);

		ret = parse_line(line, key, value);

		/* if it's a comment, ignore line */
		if (!ret) {
			if (line) free(line);
			line = NULL;
			continue;
		}
		
		for (i = 0; i < optlen; i++) {
			if (!strncasecmp(confoptions[i], key, strlen(confoptions[i]))) {
				if (*confvalues[i]) {
					/* redefinition of a key */
					if (line) free(line);
					return -3;
				}
				strncpy(confvalues[i], value, strlen(value));
				break;			
			}
		}
		if (i == optlen) {
			/* invalid configuration option */
			if (line) free(line);
			return -2;
		}

		if (line) free(line);
		line = NULL;
	}

	if (line) free(line);

	fclose(fp);

	return 0;
}

struct simopts* load_configuration(char *fname) {
	int i;
	int ret;
	struct simopts *retval;

	char **confvalues;
	char *confoptions[] = {
		"attacker_probability", 
		"defection_probability",
		"managed_ap_probability",
		"fake_group_size",
		"terrainX",
		"terrainY",
		"num_clients",
		"num_ap",
		"range",
		"beta",
		"num_rounds",
		"iterations",
		NULL
	};

	/* read configuration */
	confvalues = (char **)malloc(12 * sizeof(char*));
	for (i = 0; i < 12; i++) {
		confvalues[i] = (char*)malloc(80);
		memset(confvalues[i], 0, 80);
	}
	ret = parse_conffile(fname, (char**)confoptions, confvalues, 12);
	if (ret == -1) {
		fprintf(stderr, "Error opening configuration file\n");
		exit(1);
	}

	if (ret == -2) {
		fprintf(stderr, "Invalid configuration option\n");
		exit(1);
	}

	if (ret == -3) {
		fprintf(stderr, "Duplicate configuration option\n");
		exit(1);
	}

	retval = (struct simopts *)malloc(sizeof(struct simopts));
	memset(retval, 0, sizeof(struct simopts));

	retval->as.attacker_ratio = strtof(confvalues[0], NULL);
	if (errno) {
		retval->as.attacker_ratio = DEF_ATTACKER_PROB;
	}

	retval->as.defection_probability = strtof(confvalues[1], NULL);
	if (errno) {
		retval->as.defection_probability = DEF_DEFECTION_PROB;
	}

	retval->as.managed_ap_ratio = strtof(confvalues[2], NULL);
	if (errno) {
		retval->as.managed_ap_ratio = DEF_MANAGED_AP_PROB;
	}

	retval->as.fake_group_size = strtol(confvalues[3], NULL, 10);
	if (errno) {
		retval->as.fake_group_size = DEF_FAKE_GROUP_SIZE;
	}

	retval->terrainX = strtof(confvalues[4], NULL);
	if (errno) {
		retval->terrainX = DEF_TERRAIN_X;
	}

	retval->terrainY = strtof(confvalues[5], NULL);
	if (errno) {
		retval->terrainY = DEF_TERRAIN_Y;
	}

	retval->num_clients = strtol(confvalues[6], NULL, 10);
	if (errno) {
		retval->num_clients = DEF_NUM_CLIENTS;
	}

	retval->num_aps = strtol(confvalues[7], NULL, 10);
	if (errno) {
		retval->num_aps = DEF_NUM_APS;
	}

	retval->range = strtof(confvalues[8], NULL);
	if (errno) {
		retval->range = DEF_RANGE;
	}

	retval->beta = strtof(confvalues[9], NULL);
	if (errno) {
		retval->beta = DEF_BETA;
	}

	retval->num_rounds = strtol(confvalues[10], NULL, 10);
	if (errno) {
		retval->num_rounds = DEF_NUM_ROUNDS;
	}

	retval->iterations = strtol(confvalues[11], NULL, 10);
	if (errno) {
		retval->iterations = DEF_ITERATIONS;
	}

	for (i = 0; i < 12; i++) {
		free(confvalues[i]);
	}
	free(confvalues);

	return retval;
}


