/**
 * simulator.c -- Implementation of the core simulator functionality.
 * 
 * Copyright (C) 2011 Pantelis A. Frangoudis <pfrag@aueb.gr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "simulator.h"

int checked = -1;
int _type1 = 0;
int _type2 = 0;

int is_type1(AP *ap1, AP *ap2, double range) {
	if (sqrt(pow(ap1->x - ap2->x, 2) +  pow(ap1->y - ap2->y, 2)) <= range) {
		return 1;
	}
	else {
		return 0;
	}
}

double *reputation_histogram() {
	int i = 0;
	int num_attackers = 0;
	int num_truthful = 0;

	for (i = 0; i < num_clients; i++) {
		if (clients[i]->defection_probability == 0) {
			fprintf(stderr, "%f\t", clients[i]->r);
		}		
	}
	fprintf(stderr, "\n");
}

int poisson(double L, double x, double y) {
	double p = 1.0;
	int k = 0;
	double T = exp(-1.0*L*x*y);
	double u;

	do {
		k++;
		u = rand_double(1.0);
		if (u == 0.0) continue;
		p *= u;
	} while (p > T);

	return k - 1;
}

/**
 * Terrain initialization
 */
void init_simulator(double tX, double tY, int nc, int nap, double r, int scenario, attack_scenario *as, int sr, double b) {
	//init PRNG
	//generate seed from /dev/urandom

	BETA = b;
	simrounds = sr;

	dsfmt_init_gen_rand(&RNG, seedme());

	//list of fake APs
	fake_aps = NULL;

	//terrain configuration
	terrainX = tX;
	terrainY = tY;
	num_clients = nc;
	
	//grid spacing -- may not use it
	double d = 0;

	//check whether some kind of "rounding" needs to take place with the AP number
	//in case we're working on a grid topology
	if (scenario == 4) {
		num_aps = (int)sqrt(nap)*(int)sqrt(nap);
		d = terrainX/(sqrt(2) + sqrt(num_aps) - 1);
		range = sqrt(2.0)/2.0*d;
	}
	else {
		num_aps = nap;
		range = r;
	}
/*
	if (scenario != 4) {
		//9/6/2011: poisson stuff
		num_aps = poisson(((double)nap)/(terrainX*terrainY), terrainX, terrainY);
		num_clients = poisson(((double)nc)/(terrainX*terrainY), terrainX, terrainY);
	}
*/	
	next_fake_ap_id = num_aps;
	_num_edges = 0;

	//TODO: check fixed size
	fake_group_size = as->fake_group_size;

	int i = 0;
	int j = 0;

	switch (scenario) {
		case 0:
			ap_edges = 0;
			ap_reports = 0;
			client_reports = 1;
			break;
		case 1:
			ap_edges = 1;
			ap_reports = 0;
			client_reports = 1;
			break;
		case 2:
			ap_edges = 1;
			ap_reports = 1;
			client_reports = 1;
			break;
		case 3:
			ap_edges = 1;
			ap_reports = 1;
			client_reports = 0;
			break;
		case 4:
			ap_edges = 0;
			ap_reports = 0;
			client_reports = 1;
			break;
		default:
			ap_edges = 1;
			ap_reports = 0;
			client_reports = 1;
			break;
	}

	//initialize terrain blocks
#ifdef USE_BLOCKS
	blocks = (Block*)malloc( ((int)(terrainX/range) + 1) * ((int)(terrainY/range) + 1) * sizeof(Block) );
	memset(blocks, 0, ((int)(terrainX/range) + 1) * ((int)(terrainY/range) + 1) * sizeof(Block));
#endif
	double *coords = NULL;

	aps = (AP**)malloc(num_aps*sizeof(AP*));
	memset(aps, 0, num_aps*sizeof(AP*));

	clients = (Client**)malloc(num_clients * sizeof(Client*));
	memset(clients, 0, sizeof(num_clients * sizeof(Client*)));

	//place APs
	MEASURE_INIT;

	MEASURE_START;

	if (scenario != 4) {
		AP *newap;
		for (i = 0; i < num_aps; i++) {
			//alloc space for new AP
			newap = (AP*)malloc(sizeof(AP));
			memset(newap, 0, sizeof(AP));
	
			newap->id = i;
	
			//place AP randomly
			coords = random_ap_coordinates();
			newap->x = coords[0];
			newap->y = coords[1];
			newap->assoc_list = NULL;
			newap->scan_list = NULL;
			newap->num_assoc = 0;
			newap->scan_list_size = 0;
	
			free(coords);
	
			aps[i] = newap;

			ap_set_behaviour(aps[i], as);
#ifdef USE_BLOCKS
			add_ap_to_block(aps[i]);
#endif
		}
	}
	else {
		create_ap_grid(d);
	}
	MEASURE_END("Placing APs");

	//AP scan
/* 10/6: should we scan at init_simulator? better only do it at init_round */
/*
	if (scenario != 4) {
		for (i = 0; i < num_aps; i++) {
			ap_scan(aps[i]);
		}
	}
*/

	//init clients [but do not place them]
	MEASURE_START;
	Client *client;
	for (i = 0; i < num_clients; i++) {
		//alloc space for new client
		client = (Client*)malloc(sizeof(Client));
		memset(client, 0, sizeof(Client));
		
		client->id = i;

		client->fake_list_size = 0;
		client->scan_list_size = 0;
		client->scan_list = NULL;
		client->assoc_ap = NULL;
		client->fake_list = NULL;
		set_strategy(client, as);

		clients[i] = client;
	}
	MEASURE_END("Placing Clients");

	filter_threshold = 1.0;

	MEASURE_START;
	//IG = init_ig(aps, num_aps);
	MEASURE_END("Initialising IG");
}

/////////////////
/**
 * Round initialization:
 * - Flush client and AP state
 * - Dealloc/Realloc fake_aps
 * - Place clients randomly [teleportation]
 * - Clients scan/assoc with AP
 * - Each potential attacker picks a strategy for this round [depending on P_defect]
 */
void init_round() {
	//list of fake APs
	fake_aps = NULL;

	next_fake_ap_id = num_aps;
	_num_edges = 0;
	_type1 = 0;
	_type2 = 0;

	//TODO: check fixed size
	//fake_group_size = 3;

	int i = 0;
	int j = 0;

	double *coords = NULL;

	MEASURE_INIT;

	//AP scan -- IG needs to be re-initialized
	for (i = 0; i < num_aps; i++) {
		ap_scan(aps[i]);
	}

	MEASURE_START;
	IG = init_ig(aps, num_aps);
	MEASURE_END("Initialising IG");

	//place clients
	MEASURE_START;
	Client *client;
	for (i = 0; i < num_clients; i++) {
		//place client randomly
		coords = random_client_coordinates();
		clients[i]->x = coords[0];
		clients[i]->y = coords[1];
		free(coords);
		clients[i]->fake_list_size = 0;
		clients[i]->scan_list_size = 0;
		clients[i]->scan_list = NULL;
		clients[i]->assoc_ap = NULL;
		clients[i]->fake_list = NULL;
		clients[i]->audited_edges = 0;
		clients[i]->reported_edges = 0;
		clients[i]->reportedT1 = 0;
		clients[i]->reportedT2 = 0;

		//set_strategy(clients[i], as);

		//scan & associate
		associate(clients[i]);
	}
	MEASURE_END("Placing Clients");
}

/**
 * Called at the end of each round:
 * - Cleanup AP and client state
 * - Resets IG
 */
void end_round() {
	int i = 0;
	for (i = 0; i < num_clients; i++) {
		flush_client_info(clients[i]);
	}

	for (i = 0; i < num_aps; i++) {
		flush_ap_info(aps[i]);
	}

	//free fake aps
	if (fake_aps != NULL) {
		for (i = 0; i < next_fake_ap_id - num_aps; i++) {
			free_ap(fake_aps[i]);
		}
		free(fake_aps);
		fake_aps = NULL;
	}

	//free real IG
	free_ig(IG, num_aps);
	IG = NULL;
//fprintf(stderr, "Type1/Type2 = %d/%d, %f\n", _type1, _type2, ((float)_type1) / (float)(_type1 + _type2));
}
///////////////////

void init_simulator_from_file(double tX, double tY, int nc, char *fname, double r, int scenario, attack_scenario *as, int sr, double b) {
	//init PRNG
	//generate seed from /dev/urandom
	simrounds = sr;
	BETA = b;

	dsfmt_init_gen_rand(&RNG, seedme());

	//list of fake APs
	fake_aps = NULL;

	//grid spacing -- may not use it
	double d = 0;

	range = r;

	//terrain configuration
	terrainX = tX;
	terrainY = tY;
	num_clients = nc;
	
	_num_edges = 0;

	//TODO: check fixed size
	fake_group_size = as->fake_group_size;

	int i = 0;
	int j = 0;

	switch (scenario) {
		case 0:
			ap_edges = 0;
			ap_reports = 0;
			client_reports = 1;
			break;
		case 1:
			ap_edges = 1;
			ap_reports = 0;
			client_reports = 1;
			break;
		case 2:
			ap_edges = 1;
			ap_reports = 1;
			client_reports = 1;
			break;
		case 3:
			ap_edges = 1;
			ap_reports = 1;
			client_reports = 0;
			break;
		default:
			ap_edges = 1;
			ap_reports = 0;
			client_reports = 1;
			break;
	}

	//initialize terrain blocks
#ifdef USE_BLOCKS
	blocks = (Block*)malloc( ((int)(terrainX/range) + 1) * ((int)(terrainY/range) + 1) * sizeof(Block) );
	memset(blocks, 0, ((int)(terrainX/range) + 1) * ((int)(terrainY/range) + 1) * sizeof(Block));
#endif
	double *coords = NULL;

	clients = (Client**)malloc(num_clients * sizeof(Client*));
	if (num_clients > 0) {
		memset(clients, 0, sizeof(num_clients * sizeof(Client*)));
	}

	//place APs
	MEASURE_INIT;

	MEASURE_START;
	load_topology(fname);
	MEASURE_END("Loading topology from file");

	//AP scan
	for (i = 0; i < num_aps; i++) {
		ap_scan(aps[i]);
	}

	//place clients
	MEASURE_START;
	for (i = 0; i < nc; i++) {
		//alloc space for new AP
		clients[i] = (Client*)malloc(sizeof(Client));
		memset(clients[i], 0, sizeof(Client));
		
		clients[i]->id = i;

		clients[i]->fake_list_size = 0;
		clients[i]->scan_list_size = 0;
		clients[i]->scan_list = NULL;
		clients[i]->assoc_ap = NULL;
		clients[i]->fake_list = NULL;
		set_strategy(clients[i], as);
	}
	MEASURE_END("Placing Clients");

	filter_threshold = 1.0;

	MEASURE_START;
	IG = init_ig(aps, num_aps);
	MEASURE_END("Initialising IG");
}

/**
 * Generate random AP coordinates
 */
double *random_ap_coordinates() {
	double *retval = (double*)malloc(2*sizeof(double));
	retval[0] = rand_double(terrainX);
	retval[1] = rand_double(terrainY);

	return retval;
}

/**
 * Generate random client coordinates
 */
double *random_client_coordinates() {
	double *retval = (double*)malloc(2*sizeof(double));
	retval[0] = rand_double(terrainX);
	retval[1] = rand_double(terrainY);

	return retval;
}

/**
 * Returns a random double in the [0, limit) range
 */
double rand_double(double limit) {
	return dsfmt_genrand_close_open(&RNG) * limit;
}

/**
 * Free allocated memory
 */
void term_simulator() {
	int i = 0;
	for (i = 0; i < num_aps; i++) {
		free_ap(aps[i]);
	}
	free(aps);

	//free fake aps
	if (fake_aps != NULL) {
		for (i = 0; i < next_fake_ap_id - num_aps; i++) {
			free_ap(fake_aps[i]);
		}
		free(fake_aps);
	}

	for (i = 0; i < num_clients; i++) {
		free_client(clients[i]);
	}
	free(clients);

	//free real IG
	if (IG) {
		free_ig(IG, num_aps);
	}

	//free reported IG
	//free_ig(reportedIG, next_fake_ap_id);

	//free blocks
#ifdef USE_BLOCKS
	free_blocks();
#endif
}

/**
 * Generates real client-based interference graph
 * CHANGED: allagi ston tropo prosthikis akmwn. Pleon, ypologizoume
 * ap'eytheias kai ta vari twn akmwn. Episis, kratame tin timi
 * _num_edges wste na ti xrhsimopoihsoume otan ypologizoume pithanotita detection
 */
void create_interference_graph() {
	int i = 0;
	int j = 0;
	int k = 0;
	int from = 0;
	int to = 0;
	Client *c;
	double weight;
	int roamers = 0;
	int newedge = 0;
	//for each client, check its scan list and add edges to IG
	for (k = 0; k < num_clients; k++) {
		c = clients[k];

		//Client not associated with anyone --> no report
		if (c->assoc_ap == NULL) {
			continue;
		}

		weight = c->r;

		for (i = 0; i < c->scan_list_size; i++) {
			from = c->scan_list[i]->id;
			for (j = i; j < c->scan_list_size; j++) {
				if (i == j) continue;
				to = c->scan_list[j]->id;

				/* 6/5/11: if both unmanaged, edge ignored */
				if (c->scan_list[i]->is_managed || c->scan_list[j]->is_managed) {
					newedge = add_edge(IG, from, to, weight, !c->will_attack);
					_num_edges += newedge;
					newedge	= add_edge(IG, to, from, weight, !c->will_attack); 
					_num_edges += newedge;

					if (newedge) {
						if (sqrt( pow(c->scan_list[i]->x - c->scan_list[j]->x, 2) + pow(c->scan_list[i]->y - c->scan_list[j]->y, 2) ) <= range) {
							_type1++;
						}
						else {
							_type2++;
						}
					}
				}
			}
		}
	}
}

/**
 * Generates real interference graph including interference among APs [AP-reported]
 * changed: ap report weight >= 1
 */
void create_full_interference_graph() {
	int i = 0;
	int j = 0;
	int k = 0;
	int from = 0;
	int to = 0;
	Client *c;
	int newedge = 0;

	//client-based edges
	create_interference_graph();

	//the same for each AP
	for (k = 0; k < num_aps; k++) {
		if (aps[k]->scan_list_size <= 0) continue;

		for (i = 0; i < aps[k]->scan_list_size; i++) {
			from = aps[k]->scan_list[i]->id;

			/* 6/5/2011: if both unmanaged, edge ignored */
			if ( (aps[k]->is_managed || aps[k]->scan_list[i]->is_managed) && (aps[i]->id != aps[k]->scan_list[i]->id) ) {
				newedge = add_edge(IG, from, aps[k]->id, AP_WEIGHT, 0);
				_num_edges += newedge;
				newedge = add_edge(IG, aps[k]->id, from, AP_WEIGHT, 0);
				_num_edges += newedge;
				if (newedge) {
					if (sqrt( pow(aps[k]->x - aps[k]->scan_list[i]->x, 2) + pow(aps[k]->y - aps[k]->scan_list[i]->y, 2) ) <= range ) {
						_type1++;
					}
					else {
						_type2++;
					}
				}
			}

			for (j = i; j < aps[k]->scan_list_size; j++) {
				if (i == j) continue;
				to = aps[k]->scan_list[j]->id;
				if ( (aps[k]->scan_list[i]->is_managed || aps[k]->scan_list[j]->is_managed) && (aps[k]->scan_list[i]->id != aps[k]->scan_list[j]->id) ) {
					newedge = add_edge(IG, from, to, AP_WEIGHT, 0);
					_num_edges += newedge;
					newedge = add_edge(IG, to, from, AP_WEIGHT, 0);
					_num_edges += newedge;

					if (newedge) {
						if (sqrt( pow(aps[k]->scan_list[i]->x - aps[k]->scan_list[j]->x, 2) + pow(aps[k]->scan_list[i]->y - aps[k]->scan_list[j]->y, 2) ) <= range ) {
							_type1++;
						}
						else {
							_type2++;
						}
					}
				}
			}
		}
		
	}
}

/**
 * Filters edges from the reported IG
 * CHANGED: pleon to filtering ginetai panw sto original IG
 */
void filter() {
	int i = 0;
	int j = 0;
	int k = 0;
	int edge_weight = 0;
	Edge *cur;
	Edge *prev;

	for (i = 0; i < next_fake_ap_id; i++) {
		cur = IG[i].neighbors;
		prev = NULL;
		while (cur != NULL) {
			if (cur->weight < filter_threshold) {
				//first remove the edge's "mirror" edge
				//int to = cur->to->ap->id;
				//remove_edge(IG, to, i);

				//If edge is removed, we have to "punish"
				//all clients who reported it
				/* 6/5/11 */
				if (IG[i].ap->is_managed || cur->to->ap->is_managed) {
					for (k = 0; k < cur->num_reporters; k++) {
						update_score_info(cur->reporters[k], 0);
					}
				}

//TOCHECK!!!!
				if (prev == NULL) {
					IG[i].neighbors = cur->next;
					if (cur->reporters) {
						free(cur->reporters);
					}
					free(cur);
					cur = IG[i].neighbors;
				}
				else {
					prev->next = cur->next;
					if (cur->reporters) {
						free(cur->reporters);
					}
					free(cur);
					cur = prev->next;
				}
			}
			else {
				/* 6/5/11 */
				if (IG[i].ap->is_managed || cur->to->ap->is_managed) {
					for (k = 0; k < cur->num_reporters; k++) {
						update_score_info(cur->reporters[k], 1);
					}
				}
				prev = cur;
				cur = cur->next;
			}
		}
	}
}

/**
 * Compare real vs (filtered) reported IG and count edges found
 * CHANGED: Anti na sygkrinw 2 graphs, kanw to ekshs:
 * Metraw poses akmes exei to filtered IG kai to sygkrinw me to _num_edges
 * [prepei na diairesw kai /2 giati kathe edge tin exw dipli [from <--> to]
 */
double calculate_detection_probability() {
	Edge *cur;
	Edge *prev;
	int edges_found = 0;
	int i = 0;
	int t2found = 0;
	for (i = 0; i < next_fake_ap_id; i++) {
		cur = IG[i].neighbors;
		while (cur != NULL) {
			/* 6/5/2011 */
			if (IG[i].ap->is_managed || cur->to->ap->is_managed) {
				if (sqrt( pow(IG[i].ap->x - cur->to->ap->x, 2) + pow(IG[i].ap->y - cur->to->ap->y, 2) ) > range) {
					t2found++;
				}
				edges_found++;
			}
			cur = cur->next;
		}
	}
//printf("\n\n%d/%d\n\n", edges_found, _num_edges);
	if (_num_edges == 0) {
		return -1;
	}
	else {
/*
fprintf(stderr, "\n\nType 2 discovered: %f, %f\n", ((float)t2found)/2.0, ((float)t2found)/(float)edges_found);
fprintf(stderr, "DP: %f\n", edges_found/(double)_num_edges);
*/
		return edges_found/((double)_num_edges);
	}

}

/**
 * Carry out simulation
 * CHANGED!:
 * 1. Create full IG, counting edges and considering report "weights"
 * 2. Clients (and/or APs) report & their reports change the real IG only if !truthful
 * 3. Filter (as before, but in the real IG this time)
 * 4. Calculate DP in a different way
 *
 * This time, instead of attacking, attackers just do nothing
 */
struct round_stats* simulate() {
	int i = 0;
	int j = 0;
	Client *c;

	struct round_stats *retval = (struct round_stats*)malloc(simrounds * sizeof(struct round_stats));

	for (i = 0; i < simrounds; i++) {
		fprintf(stderr, "Round %d\r", i + 1);
		init_round();

		for (j = 0; j < num_clients; j++) {
			//decide if client will attack in this round
			attack(clients[j]);
		}

		//generate real IG
		if (!ap_edges) {
			create_interference_graph();
		}
		else {
			create_full_interference_graph();
		}

		//clients report
		/* 15/5: report calls relax_edge */
		for (j = 0; j < num_clients; j++) {
			report(clients[j], client_reports);
		}

		for (j = 0; j < num_aps; j++) {
			//[was: no AP reports] managed APs should report [5/5/2011]
			ap_report(aps[j], ap_edges, ap_reports);
		}

		/* dirty hack: get stats before filtering (we only record reputations) */
		struct round_stats *at_round_start = get_round_stats();

		//filter IG depending on scenario
		/* truthful user scores are updated inside filter() */
		filter();

		//calculate detection probability
		double dp = calculate_detection_probability();

		//discount reputation for attackers
		int x = 0;
		for (j = 0; j < num_clients; j++) {
			if (clients[j]->assoc_ap != NULL && clients[j]->assoc_ap->is_managed && clients[j]->will_attack) {
				//maybe fgs + 1?? -- not important at the moment
				//clients[j]->r = pow(clients[j]->r, (double)fake_group_size*(fake_group_size - 1.0) / 2.0);
				for (x = 0; x < fake_group_size*(fake_group_size - 1.0) / 2.0; x++) {
					update_score_info(clients[j], 0);
				}
			}
		}

		for (j = 0; j < num_clients; j++) {
			update_reputation(clients[j]);
		}

		struct round_stats *rs = get_round_stats();
		retval[i].dp = dp;
		retval[i].truthful_rep = at_round_start->truthful_rep;
		retval[i].attacker_rep = at_round_start->attacker_rep;
		retval[i].isolated = rs->isolated;
		retval[i].truthful_score = rs->truthful_score;
		retval[i].attacker_score = rs->attacker_score;
		retval[i].all_rep = at_round_start->all_rep;
		retval[i].all_score = rs->all_score;

		if (rs) free(rs);
		if (at_round_start) free(at_round_start);
//print_graph();

//		reputation_histogram();
/*
		retval[i][1] = avgrep[0]; //rep attackers
		retval[i][2] = avgrep[1]; //rep truthful
		retval[i][3] = avgrep[2]; //score (truthful)
		retval[i][4] = avgrep[3]; //P(isolated)
		retval[i][5] = avgrep[4]; //score across all
		retval[i][6] = avgrep[5]; //rep across all
*/
		//fprintf(stderr, "%d\t%f\t%f\t%f\t%f\n", i + 1, dp, avgrep[0], avgrep[1], avgrep[2]);
/*
int type1 = 0;
int type2 = 0;
int k = 0;
for (i = 0; i < num_clients; i++) {
	type1 = 0;
	type2 = 0;
	for (j = 0; j < clients[i]->scan_list_size; j++) {
		for (k = j + 1; k < clients[i]->scan_list_size; k++) {
			if (is_type1(clients[i]->scan_list[j], clients[i]->scan_list[k], range)) type1++;
			else type2++;
		}
	}
	fprintf(stderr, "client[%d], type1: %f, type2: %f\n", i, ((float)type1)/(float)(type1+type2), ((float)type2)/(float)(type1+type2));
}
*/
		if (i == 8) {
//			reputation_histogram();
		}
		end_round();
	}

	return retval;
}

unsigned int seedme() {
	unsigned int theseed;
	FILE *fp = fopen("/dev/urandom", "r");
	if (!fp) {
		theseed = (unsigned int)time(NULL);
	}
	else {
		fread(&theseed, 1, sizeof(int), fp);
		fclose(fp);
	}
	return theseed;
}

int get_block(double x, double y) {
	int blocks_per_row = terrainX/range;
	
	int bx = x/range;
	int by = y/range;
	
	return by*blocks_per_row + bx;
}

void add_ap_to_block(AP *ap) {
	int block_id = get_block(ap->x, ap->y);

	if (blocks[block_id].apnum == 0) {
		blocks[block_id].aps = (AP**)malloc(sizeof(AP*));
	}
	else {
		blocks[block_id].aps = (AP**)realloc(blocks[block_id].aps, (blocks[block_id].apnum  + 1)*sizeof(AP*) );
	}
	blocks[block_id].aps[blocks[block_id].apnum] = ap;
	blocks[block_id].apnum++;
}

void free_blocks() {
	int i = 0;

	if (!blocks) return;

	for (i = 0; i < ((int)(terrainX/range) + 1) * ((int)(terrainY/range) + 1); i++) {
		if (blocks[i].apnum > 0) {
			free(blocks[i].aps);
		}
	}

	free(blocks);
}

void create_ap_grid(double d) {	
	double curX = range;
	double curY = range;

	AP *newap;
	int i = 0;

        for (i = 0; i < num_aps; i++) {
		if (curX > terrainX) {
			curY += d;
			curX = range;
		}

		newap = (AP*)malloc(sizeof(AP));
		memset(newap, 0, sizeof(AP));

		newap->id = i;

		newap->x = curX;
		newap->y = curY;
		newap->assoc_list = NULL;
		newap->scan_list = NULL;
		newap->num_assoc = 0;
		newap->scan_list_size = 0;
	
		aps[i] = newap;
#ifdef USE_BLOCKS
		add_ap_to_block(aps[i]);
#endif
		curX +=d;
        }
}

void load_topology(char *fname) {
	FILE *fp = fopen(fname, "r");

	AP *newap;
	int i = 0;

	char line[80];
	memset (line, 0, 80);

	//number of lines (each line --> an AP);
	int lines = 0;

	while (fgets(line, 80, fp)) {
		char *pos = index(line, ' ');
		if (!pos) continue;
		lines++;
	}
	
	num_aps = lines;

	aps = (AP**)malloc(num_aps*sizeof(AP*));
	memset(aps, 0, num_aps*sizeof(AP*));


	fseek(fp, 0L, SEEK_SET);

	while (fgets(line, 80, fp)) {
		char *pos = index(line, ' ');
		if (!pos) continue;
		
		char x[80];memset(x, 0, 80);
		char y[80];memset(y, 0, 80);

		strncpy(x, line, (pos - line) + 1);
		strcpy(y, pos);

		newap = (AP*)malloc(sizeof(AP));
		memset(newap, 0, sizeof(AP));

		newap->id = i;

		newap->x =  (double)strtod(x, NULL);
		newap->y = (double)strtod(y, NULL);
		newap->assoc_list = NULL;
		newap->scan_list = NULL;
		newap->num_assoc = 0;
		newap->scan_list_size = 0;
	
		aps[i] = newap;
#ifdef USE_BLOCKS
		add_ap_to_block(aps[i]);
#endif
		i++;

	}
	fclose(fp);
}

void print_graph() {
	int i = 0;
	int edge_weight = 0;
	Edge *cur;

	for (i = 0; i < num_aps; i++) {

		fprintf(stderr, "Node %d: ", IG[i].ap->id);
		cur = IG[i].neighbors;
		while (cur != NULL) {
			fprintf(stderr, "%d[%f]\t", cur->to->ap->id, cur->weight);
			cur = cur->next;
		}
		fprintf(stderr, "\n");
	}
}

/**
 * success == {0,1}
 */
void update_score_info(Client *c, int success) {
	//instead of updating reputation per edge,
	//we keep track of user's reported/audited edges
	//and calculate score at the end of a round
	//c->r = BETA*c->r + success*(1.0 - BETA);
//	c->reported_edges++;
	c->audited_edges += success;
}

/**
 * success == {0,1}
 */
void update_reputation(Client *c) {
	double round_score = 0.0;

	/* 5/5/2011 */
	if (c->assoc_ap == NULL || !c->assoc_ap->is_managed || c->scan_list_size == 1) {
	//if (c->assoc_ap == NULL || !c->assoc_ap->is_managed) {
		/* client is "isolated". his reputation remains as is */
		return;
	}
	
	round_score = ((double)c->audited_edges)/((double)c->reported_edges + 0.00001);
	if (c->will_attack) {
		round_score = 0;
	}

	else {
//		round_score = calculate_detection_probability();
	}

	c->r = BETA*c->r + round_score*(1.0 - BETA);

}

/* Returns this tuple: [avg attacker reputation, avg truthful reputation, avg truthful score] */
struct round_stats *get_round_stats() {
	int i = 0;
	double sum_rep_attackers = 0.0;
	double sum_rep_truthful = 0.0;
	double sum_score_truthful = 0.0;
	double sum_score_attackers = 0.0;

	int num_attackers = 0;
	int num_truthful = 0;
	int to_divide_by_a = 0;
	int to_divide_by_t = 0;
	double isolated = 0.0;
	double score_for_all = 0.0;
	double rep_for_all = 0.0;

	double sum_reported_type1 = 0.0;
	double sum_reported_type2 = 0.0;

	struct round_stats *retval;

	for (i = 0; i < num_clients; i++) {
		if (clients[i]->defection_probability > 0.0) {
			sum_rep_attackers += clients[i]->r;
			if (clients[i]->assoc_ap != NULL && clients[i]->assoc_ap->is_managed && !clients[i]->will_attack && clients[i]->scan_list_size > 1) {
			//if (clients[i]->assoc_ap != NULL && clients[i]->assoc_ap->is_managed && !clients[i]->will_attack) {

				sum_score_attackers += ((double)clients[i]->audited_edges) / ((double)clients[i]->reported_edges + 0.00001);
//				sum_reported_edges_A += (double)clients[i]->reported_edges;

//				if ((double)clients[i]->reported_edges) {
//					fprintf(stderr, "attacker %d: %d/%d=%f\n", i, clients[i]->audited_edges, clients[i]->reported_edges, ((double)clients[i]->audited_edges) / ((double)clients[i]->reported_edges + 0.00001));
//				}

				to_divide_by_a++;
			}
			num_attackers++;
		}
		else {
			/* 6/5/2011: calculate avg score only for non-isolated clients */
			if (clients[i]->assoc_ap != NULL && clients[i]->assoc_ap->is_managed && clients[i]->scan_list_size > 1) {
			//if (clients[i]->assoc_ap != NULL && clients[i]->assoc_ap->is_managed) {
				sum_score_truthful += ((double)clients[i]->audited_edges) / ((double)clients[i]->reported_edges + 0.00001);
//sum_score_truthful += DP;
				to_divide_by_t++;

//				sum_reported_edges_T += (double)clients[i]->reported_edges;
			}
			sum_rep_truthful += clients[i]->r;
			num_truthful++;
		}

		//////////////
		if (clients[i]->assoc_ap == NULL || !clients[i]->assoc_ap->is_managed || clients[i]->scan_list_size == 1) {
		//if (clients[i]->assoc_ap == NULL || !clients[i]->assoc_ap->is_managed) {
			isolated++;
		}
		else {
			sum_reported_type1 += clients[i]->reportedT1;
			sum_reported_type2 += clients[i]->reportedT2;
		}

		rep_for_all += clients[i]->r;
	}

//fprintf(stderr, "\n Type-1: %f\tType-2: %f\n\n", sum_reported_type1 / (double)(to_divide_by_t + to_divide_by_a), sum_reported_type2 / (double)(to_divide_by_t + to_divide_by_a));

	retval = (struct round_stats*)malloc(sizeof(struct round_stats));

	retval->attacker_rep = sum_rep_attackers / (double)num_attackers;
	retval->truthful_rep = sum_rep_truthful / (double)num_truthful;
	retval->truthful_score = sum_score_truthful / (double)to_divide_by_t;
	retval->attacker_score = sum_score_attackers / (double)to_divide_by_a;
//fprintf(stderr, "%f / %f = %f\n\n", sum_score_attackers, (double)to_divide_by_a, sum_score_attackers / (double)to_divide_by_a);
	retval->isolated = isolated/(double)num_clients;
	retval->all_score = (sum_score_truthful + sum_score_attackers) / (double)(to_divide_by_t + to_divide_by_a);
	retval->all_rep = rep_for_all / (double)num_clients;

	return retval;
}


