/**
 * simulator.h -- Implementation of the core simulator functionality.
 * 
 * Copyright (C) 2011 Pantelis A. Frangoudis <pfrag@aueb.gr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef _SIMULATOR_H_
#define _SIMULATOR_H_

#include "types.h"
#include "client.h"
#include "accesspoint.h"
#include "ig.h"
#include "dSFMT.h"
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#ifdef PERFMSR

#define MEASURE_INIT \
        struct timeval tv_start; \
        struct timeval tv_end; \
        long diff = 0L; \

#define MEASURE_START \
        memset(&tv_start, 0, sizeof(tv_start)); \
        gettimeofday(&tv_start, NULL); \

#define MEASURE_END(str) \
        memset(&tv_end, 0, sizeof(tv_end)); \
        gettimeofday(&tv_end, NULL); \
        if (str) { \
                fprintf(stderr, "%s took %ld usec\n", str, tv_end.tv_sec*1000000 + tv_end.tv_usec - tv_start.tv_sec*1000000 - tv_start.tv_usec); \
        } \

#else

#define MEASURE_INIT \

#define MEASURE_START \

#define MEASURE_END(str) \

#endif

/**
 * Nodes
 */
Client **clients;
AP **aps;
AP **fake_aps;

int next_fake_ap_id;
int fake_group_size;

/**
 * Terrain size
 */
double terrainX;
double terrainY;

/**
 * Terrain blocks
 */
Block *blocks;

/**
 * Transmission range -- assumed fixed
 */
double range;

/**
 * Number of nodes
 */
int num_aps;
int num_clients;

/**
 * PRNG -- Mersenne Twister
 */
dsfmt_t RNG;

/**
 * Interference Graph
 */
//double **IG;
//double **reportedIG;
Node *IG;
Node *reportedIG;

double filter_threshold;

/**
 * CHANGE: we need to keep track of the number of real IG edges
 */
int _num_edges;

int ap_edges;
int ap_reports;
int client_reports;

//////////////
double BETA;
int simrounds;
//////////////

/**
 * Generate random coordinates
 */
double *random_ap_coordinates();
double *random_client_coordinates();

/**
 * Random double in the [0, limit) interval
 */
double rand_double(double limit);

/**
 * Simulator initialization. Places clients and APs
 */
void init_simulator(double tX, double tY, int nc, int nap, double range, int scenario, attack_scenario *as, int simrounds, double b);
void init_simulator_from_file(double tX, double tY, int nc, char *fname, double range, int scenario, attack_scenario *as, int simrounds, double b);

void init_round();
void end_round();

/**
 * Executes simulation and returns detection probability
 */
struct round_stats* simulate();

/**
 * Simulator cleanup. Free memory...
 */
void term_simulator();

/**
 * Generates client-based real interference graph
 */
void create_interference_graph();

/**
 * Generates real interference graph including interface among APs [AP-reported]
 */
void create_full_interference_graph();

/**
 * Filters reported IG
 */
void filter();

/**
 * Returns percentage of detected edges
 */
double calculate_detection_probability();

/**
 * Prints info on client population
 */
//void population_statistics();

/**
 * Reads from /dev/urandom & seeds RNG
 */
unsigned int seedme();

/**
 * Calculates block ID. Assumes that blocks are numbered sequentially
 * Block ID == order of block in the "blocks" array
 */
int get_block(double x, double y);

/**
 * Registers an AP with a block
 */
void add_ap_to_block(AP *ap);

void free_blocks();

void create_ap_grid();

void load_topology();
void print_graph();

void update_reputation(Client *c);
void update_score_info(Client *c, int success);

struct round_stats *get_round_stats();

int is_type1(AP *ap1, AP *ap2, double range);

#endif
